# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import json
import typing
from abc import abstractmethod
from datetime import datetime, timezone
from typing import cast

import botocore.exceptions
import fsspec

import mlrun.datastore.base
from mlrun.common.schemas.model_monitoring.constants import StatsKind
from mlrun.model_monitoring.helpers import (
    get_monitoring_current_stats_data,
    get_monitoring_drift_measures_data,
    get_monitoring_stats_directory_path,
)
from mlrun.utils import logger


class ModelMonitoringStatsFile(abc.ABC):
    """
    Abstract class
    Initialize applications monitoring stats file object.
    The JSON file stores a dictionary of registered application name as key and Unix timestamp as value.
    When working with the schedules data, use this class as a context manager to read and write the data.
    """

    def __init__(self, item: mlrun.datastore.base.DataItem, file_type: str):
        self._path = item.url
        self._item = item
        self._file_type = file_type
        self._fs = cast(fsspec.AbstractFileSystem, self._item.store.filesystem)

    def create(self) -> None:
        """Create a json file with initial content - an empty dictionary"""
        logger.debug(
            f"Creating model monitoring {self._file_type} file", path=self._item.url
        )
        self._item.put(
            json.dumps(
                {
                    "data": dict(),
                    "timestamp": mlrun.utils.datetime_now().isoformat(
                        sep=" ", timespec="microseconds"
                    ),
                }
            )
        )

    def delete(self) -> None:
        """Delete json file if it exists"""
        if self._fs.exists(self._path):
            logger.debug(
                f"Deleting model monitoring {self._file_type} file", path=self._item.url
            )
            self._item.delete()
        else:
            logger.debug(
                f"Model monitoring {self._file_type} file does not exist, nothing to delete",
                path=self._item.url,
            )

    def read(self) -> tuple[dict, typing.Optional[datetime]]:
        """
        Read the stats data and timestamp saved in file
        :return: tuple[dict, str] dictionary with stats data and timestamp saved in file
        """
        try:
            content = json.loads(self._item.get().decode())
            timestamp = content.get("timestamp")
            if timestamp is not None:
                timestamp = datetime.fromisoformat(timestamp).astimezone(
                    tz=timezone.utc
                )
            return content.get("data"), timestamp
        except (
            mlrun.errors.MLRunNotFoundError,
            # Different errors are raised for S3 or local storage, see ML-8042
            botocore.exceptions.ClientError,
            FileNotFoundError,
        ) as err:
            if (
                isinstance(err, botocore.exceptions.ClientError)
                # Add a log only to "NoSuchKey" errors codes - equivalent to `FileNotFoundError`
                and err.response["Error"]["Code"] != "NoSuchKey"
            ):
                raise

            logger.warning(
                "The Stats file was not found. It should have been created "
                "as a part of the model endpoint's creation",
                path=self._path,
                error=err,
            )
            return {}, None

    def write(self, stats: dict, timestamp: datetime) -> None:
        """
        Write stats data to file overwrite the existing file
        :param stats: dictionary with the stats data
        :param timestamp: datetime object with the timestamp of last entry point for the stats calculation
        """
        content = {
            "data": stats,
            "timestamp": timestamp.isoformat(sep=" ", timespec="microseconds"),
        }
        self._item.put(json.dumps(content))

    @classmethod
    @abstractmethod
    def from_model_endpoint(
        cls, model_endpoint: mlrun.common.schemas.ModelEndpoint
    ) -> "ModelMonitoringStatsFile":
        """
        Return ModelMonitoringStatsFile child object using ModelEndpoint metadata
        :param model_endpoint: The current model endpoint to get a stats object for
        :return: ModelMonitoringStatsFile child object instance
        """
        pass


class ModelMonitoringCurrentStatsFile(ModelMonitoringStatsFile):
    def __init__(self, project: str, endpoint_id: str) -> None:
        """
        Initialize File object specific for current stats.
        :param project:         (str) Project name
        :param endpoint_id:     (str) Endpoint name
        """
        super().__init__(
            get_monitoring_current_stats_data(project, endpoint_id),
            StatsKind.CURRENT_STATS.value,
        )

    @classmethod
    def from_model_endpoint(
        cls, model_endpoint: mlrun.common.schemas.ModelEndpoint
    ) -> "ModelMonitoringCurrentStatsFile":
        return cls(
            project=model_endpoint.metadata.project,
            endpoint_id=model_endpoint.metadata.uid,
        )


class ModelMonitoringDriftMeasuresFile(ModelMonitoringStatsFile):
    def __init__(self, project: str, endpoint_id: str) -> None:
        """
        Initialize File object specific for drift measures.
        :param project:         (str) Project name
        :param endpoint_id:     (str) Endpoint name
        """
        super().__init__(
            get_monitoring_drift_measures_data(project, endpoint_id),
            StatsKind.DRIFT_MEASURES.value,
        )

    @classmethod
    def from_model_endpoint(
        cls, model_endpoint: mlrun.common.schemas.ModelEndpoint
    ) -> "ModelMonitoringDriftMeasuresFile":
        return cls(
            project=model_endpoint.metadata.project,
            endpoint_id=model_endpoint.metadata.uid,
        )


def delete_model_monitoring_stats_folder(project: str) -> None:
    """Delete the model monitoring schedules folder of the project"""
    folder = get_monitoring_stats_directory_path(project)
    fs = mlrun.datastore.store_manager.object(folder).store.filesystem
    if fs and fs.exists(folder):
        logger.debug("Deleting model monitoring stats folder", folder=folder)
        fs.rm(folder, recursive=True)
    elif fs is None:  # In-memory store
        raise mlrun.errors.MLRunValueError(
            "Cannot delete a folder without a file-system"
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import json
import sys
from typing import Callable, Optional, Union

import mlrun.errors

from .utils import SKLearnTypes


class Metric:
    """
    A metric handling class to call a metric with additional keyword arguments later during a run and log the results to
    MLRun.
    """

    def __init__(
        self,
        metric: Union[Callable, str],
        name: Optional[str] = None,
        additional_arguments: Optional[dict] = None,
        need_probabilities: bool = False,
    ):
        """
        Initialize a metric object to be used with the MLRun logger.

        :param metric:               The metric to use. Can be passed as a string of an imported function or a full
                                     module path to import from.
        :param name:                 The metric name to use for logging it to MLRun.
        :param additional_arguments: Additional arguments to pass for the metric function when calculating it.
        :param need_probabilities:   Whether this metric expects 'y_pred' to be from the 'predict_proba' method or from
                                     'predict'.
        """
        self._metric = (
            self._from_string(metric=metric) if isinstance(metric, str) else metric
        )
        self._arguments = {} if additional_arguments is None else additional_arguments
        self._need_probabilities = need_probabilities
        self._name = name if name is not None else self._get_default_name()
        self._result = None  # type: Union[float, None]

    def __call__(
        self,
        y_true: SKLearnTypes.DatasetType,
        y_pred: SKLearnTypes.DatasetType = None,
        model: SKLearnTypes.ModelType = None,
        x: SKLearnTypes.DatasetType = None,
    ) -> float:
        """
        Call the metric function on the provided y_true and y_pred values using the stored additional arguments.

        :param y_true: The ground truth values.
        :param y_pred: The model predictions.

        :return: The metric result.
        """
        # Run a prediction if 'y_pred' was not given:
        if y_pred is None:
            if model is None or x is None:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Calculating a metric requires the model's predictions / probabilities (y_pred) or the model "
                    "itself and an input (x) to run 'predict' / 'predict_proba'."
                )
            y_pred = (
                model.predict_proba(x) if self._need_probabilities else model.predict(x)
            )

        # Calculate the result and return:
        self._result = self._metric(y_true, y_pred, **self._arguments)
        return self._result

    @property
    def name(self) -> str:
        """
        Get the name of the metric.

        :return: The name of the metric.
        """
        return self._name

    @property
    def need_probabilities(self) -> bool:
        """
        Return whether this metric expects 'y_pred' to be from the 'model.predict_proba' method or from 'model.predict'.

        :return: True if probabilities are required and False if not.
        """
        return self._need_probabilities

    @property
    def result(self) -> Union[float, None]:
        """
        Get the metric result. If the metric was not calculated, None will be returned.

        :return: The metric result.
        """
        return self._result

    def _get_default_name(self) -> str:
        """
        Get the default name for this metric by the following rules:

        * If metric is a function, the function's name.
        * If metric is a callable object, the object's class name.

        :return: The metric default name.
        """
        # Function object have __name__ where classes instances have __class__.__name__:
        name = getattr(self._metric, "__name__", None)
        return name if name is not None else self._metric.__class__.__name__

    def display(self, full: bool = False):
        """
        Display the metric and its result.

        :param full: Whether to print a full display of the metric, including the metric arguments. Default: False.
        """
        result = self._result if self._result is not None else "?"
        print(f"{self._name} = {result}")
        if full:
            print(f"Arguments: {json.dumps(self._arguments, indent=4)}")

    def _repr_pretty_(self, p, cycle: bool):
        """
        A pretty representation of the metric. Will be called by the IPython kernel. This method will call the metric
        display method.

        :param p:     A RepresentationPrinter instance.
        :param cycle: If a cycle is detected to prevent infinite loop.
        """
        self.display()

    @staticmethod
    def _from_string(metric: str) -> Callable:
        """
        Look for the metric by name in the globally imported objects. If the given metric is a full module path, it will
        be imported from the path.

        :param metric: The metric name or a full module path to import the metric.

        :return: The imported metric function.

        :raise MLRunInvalidArgumentError: If the metric name was not found within the global imports.
        """
        # Check if the metric is inside a module path:
        module = None  # type: Union[ModuleType, str, None]
        if "." in metric:
            module, metric = metric.rsplit(".", 1)

        # Look for the metric in the globals dictionary (it was imported before):
        if metric in globals():
            return globals()[metric]

        # Import the metric from the given module:
        if module is not None:
            # Check if the module is already imported:
            if module in sys.modules:
                # It is already imported:
                module = sys.modules[module]
            else:
                # Import the module:
                module = importlib.import_module(module)
            imported_metric = getattr(module, metric)
            globals().update({metric: imported_metric})
            return imported_metric

        # Metric string was not provided properly:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"The metric {metric} was not found in the global imports dictionary meaning it was not "
            f"imported. In order to import it during the run, please provide the full module path to the "
            f"metric. For example: 'module.sub_module.metric' will be parsed as "
            f"from module.sub_module import metric."
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tarfile
import zipfile
from abc import ABC, abstractmethod
from pathlib import Path

import mlrun.utils

from ._supported_format import SupportedFormat


class _Archiver(ABC):
    """
    An abstract base class for an archiver - a class to manage archives of multiple files.
    """

    @classmethod
    @abstractmethod
    def create_archive(cls, directory_path: str, output_path: str) -> str:
        """
        Create an archive of all the contents in the given directory and save it to an archive file named as the
        directory in the provided output path.

        :param directory_path: The directory with the files to archive.
        :param output_path:    The output path to store the created archive file.

        :return: The created archive path.
        """
        pass

    @classmethod
    @abstractmethod
    def extract_archive(cls, archive_path: str, output_path: str) -> str:
        """
        Extract the given archive to a directory named as the archive file (without the extension) located in the
        provided output path.

        :param archive_path: The archive file to extract its contents.
        :param output_path:  The output path to extract the directory of the archive to.

        :return: The extracted contents directory path.
        """
        pass


class _ZipArchiver(_Archiver):
    """
    A static class for managing zip archives.
    """

    @classmethod
    def create_archive(cls, directory_path: str, output_path: str) -> str:
        """
        Create an archive of all the contents in the given directory and save it to an archive file named as the
        directory in the provided output path.

        :param directory_path: The directory with the files to archive.
        :param output_path:    The output path to store the created archive file.

        :return: The created archive path.
        """
        # Convert to `pathlib.Path` objects:
        directory_path = Path(directory_path)
        output_path = Path(output_path)

        # Construct the archive file path:
        archive_path = output_path / f"{directory_path.stem}.zip"

        # Archive:
        with zipfile.ZipFile(archive_path, "w") as zip_file:
            for path in directory_path.rglob("*"):
                zip_file.write(filename=path, arcname=path.relative_to(directory_path))

        return str(archive_path)

    @classmethod
    def extract_archive(cls, archive_path: str, output_path: str) -> str:
        """
        Extract the given archive to a directory named as the archive file (without the extension) located in the
        provided output path.

        :param archive_path: The archive file to extract its contents.
        :param output_path:  The output path to extract the directory of the archive to.

        :return: The extracted contents directory path.
        """
        # Convert to `pathlib.Path` objects:
        archive_path = Path(archive_path)
        output_path = Path(output_path)

        # Create the directory path, add timestamp to avoid collisions:
        directory_path = output_path / archive_path.stem
        directory_path = directory_path.with_name(
            f"{directory_path.name}-{mlrun.utils.now_date().isoformat()}"
        )
        os.makedirs(directory_path)

        # Extract:
        with zipfile.ZipFile(archive_path, "r") as zip_file:
            zip_file.extractall(directory_path)

        return str(directory_path)


class _TarArchiver(_Archiver):
    """
    A static class for managing tar archives.
    """

    # Inner class variable to note how to open a `TarFile` object for reading and writing:
    _MODE_STRING = ""

    @classmethod
    def create_archive(cls, directory_path: str, output_path: str) -> str:
        """
        Create an archive of all the contents in the given directory and save it to an archive file named as the
        directory in the provided output path.

        :param directory_path: The directory with the files to archive.
        :param output_path:    The output path to store the created archive file.

        :return: The created archive path.
        """
        # Convert to `pathlib.Path` objects:
        directory_path = Path(directory_path)
        output_path = Path(output_path)

        # Construct the archive file path:
        archive_file_extension = (
            "tar" if cls._MODE_STRING == "" else f"tar.{cls._MODE_STRING}"
        )
        archive_path = output_path / f"{directory_path.stem}.{archive_file_extension}"

        # Archive:
        with tarfile.open(archive_path, f"w:{cls._MODE_STRING}") as tar_file:
            for path in directory_path.rglob("*"):
                tar_file.add(name=path, arcname=path.relative_to(directory_path))

        return str(archive_path)

    @classmethod
    def extract_archive(cls, archive_path: str, output_path: str) -> str:
        """
        Extract the given archive to a directory named as the archive file (without the extension) located in the
        provided output path.

        :param archive_path: The archive file to extract its contents.
        :param output_path:  The output path to extract the directory of the archive to.

        :return: The extracted contents directory path.
        """
        # Convert to `pathlib.Path` objects:
        archive_path = Path(archive_path)
        output_path = Path(output_path)

        # Get the archive file name (can be constructed of multiple extensions like tar.gz so `Path.stem` won't work):
        archive_file_name = archive_path
        while archive_file_name.with_suffix(suffix="") != archive_file_name:
            archive_file_name = archive_file_name.with_suffix(suffix="")
        archive_file_name = archive_file_name.stem

        # Create the directory path:
        directory_path = output_path / archive_file_name
        os.makedirs(directory_path)

        # Extract:
        with tarfile.open(archive_path, f"r:{cls._MODE_STRING}") as tar_file:
            # use 'data' to ensure no security risks are imposed by the archive files
            # see: https://docs.python.org/3/library/tarfile.html#tarfile.TarFile.extractall
            tar_file.extractall(directory_path, filter="data")

        return str(directory_path)


class _TarGZArchiver(_TarArchiver):
    """
    A static class for managing tar.gz archives.
    """

    # Inner class variable to note how to open a `TarFile` object for reading and writing:
    _MODE_STRING = "gz"


class _TarBZ2Archiver(_TarArchiver):
    """
    A static class for managing tar.bz2 archives.
    """

    # Inner class variable to note how to open a `TarFile` object for reading and writing:
    _MODE_STRING = "bz2"


class _TarXZArchiver(_TarArchiver):
    """
    A static class for managing tar.gz archives.
    """

    # Inner class variable to note how to open a `TarFile` object for reading and writing:
    _MODE_STRING = "xz"


class ArchiveSupportedFormat(SupportedFormat[_Archiver]):
    """
    Library of archive formats (file extensions) supported by some builtin MLRun packagers.
    """

    ZIP = "zip"
    TAR = "tar"
    TAR_GZ = "tar.gz"
    TAR_BZ2 = "tar.bz2"
    TAR_XZ = "tar.xz"

    _FORMAT_HANDLERS_MAP = {
        ZIP: _ZipArchiver,
        TAR: _TarArchiver,
        TAR_GZ: _TarGZArchiver,
        TAR_BZ2: _TarBZ2Archiver,
        TAR_XZ: _TarXZArchiver,
    }



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import Optional

from mlrun.utils import get_in, update_in

# headers keys with underscore are getting ignored by werkzeug https://github.com/pallets/werkzeug/pull/2622
# to avoid conflicts with WGSI which converts all header keys to uppercase with underscores.
# more info https://github.com/benoitc/gunicorn/issues/2799, this comment can be removed once old keys are removed
event_id_key = "MLRUN-EVENT-ID"
event_path_key = "MLRUN-EVENT-PATH"


def _extract_input_data(input_path, body):
    if input_path:
        if not hasattr(body, "__getitem__"):
            raise TypeError("input_path parameter supports only dict-like event bodies")
        return get_in(body, input_path)
    return body


def _update_result_body(result_path, event_body, result):
    if result_path and event_body:
        if not hasattr(event_body, "__getitem__"):
            raise TypeError(
                "result_path parameter supports only dict-like event bodies"
            )
        update_in(event_body, result_path, result)
    else:
        event_body = result
    return event_body


class StepToDict:
    """auto serialization of graph steps to a python dictionary"""

    meta_keys = [
        "context",
        "name",
        "input_path",
        "result_path",
        "full_event",
        "kwargs",
    ]

    def to_dict(
        self,
        fields: Optional[list] = None,
        exclude: Optional[list] = None,
        strip: bool = False,
    ):
        """convert the step object to a python dictionary"""
        fields = fields or getattr(self, "_dict_fields", None)
        if not fields:
            fields = list(inspect.signature(self.__init__).parameters.keys())
        if exclude:
            fields = [field for field in fields if field not in exclude]

        args = {
            key: getattr(self, key)
            for key in fields
            if getattr(self, key, None) is not None and key not in self.meta_keys
        }
        # add storey kwargs or extra kwargs
        if "kwargs" in fields and (hasattr(self, "kwargs") or hasattr(self, "_kwargs")):
            kwargs = getattr(self, "kwargs", {}) or getattr(self, "_kwargs", {})
            for key, value in kwargs.items():
                if key not in self.meta_keys:
                    args[key] = value

        mod_name = self.__class__.__module__
        class_path = self.__class__.__qualname__
        if mod_name not in ["__main__", "builtins"]:
            class_path = f"{mod_name}.{class_path}"
        struct = {
            "class_name": class_path,
            "name": self.name
            if hasattr(self, "name") and self.name
            else self.__class__.__name__,
            "class_args": args,
        }
        if hasattr(self, "_STEP_KIND"):
            struct["kind"] = self._STEP_KIND
        if hasattr(self, "_input_path") and self._input_path is not None:
            struct["input_path"] = self._input_path
        if hasattr(self, "_result_path") and self._result_path is not None:
            struct["result_path"] = self._result_path
        if hasattr(self, "_full_event") and self._full_event:
            struct["full_event"] = self._full_event
        return struct


class MonitoringApplicationToDict(StepToDict):
    _STEP_KIND = "monitoring_application"
    meta_keys = []


class RouterToDict(StepToDict):
    _STEP_KIND = "router"

    def to_dict(
        self,
        fields: Optional[list] = None,
        exclude: Optional[list] = None,
        strip: bool = False,
    ):
        return super().to_dict(exclude=["routes"], strip=strip)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import pydantic.v1

import mlrun.common.types

from .object import ObjectStatus


class ArtifactCategories(mlrun.common.types.StrEnum):
    model = "model"
    dataset = "dataset"
    document = "document"
    llm_prompt = "llm-prompt"
    other = "other"

    # we define the link as a category to prevent import cycles, but it's not a real category
    # and should not be used as such
    link = "link"

    def to_kinds_filter(self) -> tuple[list[str], bool]:
        link_kind = ArtifactCategories.link.value

        if self.value == ArtifactCategories.model.value:
            return [ArtifactCategories.model.value, link_kind], False
        if self.value == ArtifactCategories.dataset.value:
            return [ArtifactCategories.dataset.value, link_kind], False
        if self.value == ArtifactCategories.document.value:
            return [ArtifactCategories.document.value, link_kind], False
        if self.value == ArtifactCategories.llm_prompt.value:
            return [ArtifactCategories.llm_prompt.value, link_kind], False
        if self.value == ArtifactCategories.other.value:
            return (
                [
                    ArtifactCategories.model.value,
                    ArtifactCategories.dataset.value,
                    ArtifactCategories.document.value,
                    ArtifactCategories.llm_prompt.value,
                ],
                True,
            )

    @classmethod
    def from_kind(cls, kind: str) -> "ArtifactCategories":
        if kind in [
            cls.model.value,
            cls.dataset.value,
            cls.document.value,
            cls.llm_prompt.value,
        ]:
            return cls(kind)
        return cls.other

    @staticmethod
    def all():
        """Return all applicable artifact categories"""
        return [
            ArtifactCategories.model,
            ArtifactCategories.dataset,
            ArtifactCategories.document,
            ArtifactCategories.llm_prompt,
        ]


class ArtifactIdentifier(pydantic.v1.BaseModel):
    # artifact kind
    kind: typing.Optional[str]
    key: typing.Optional[str]
    iter: typing.Optional[int]
    uid: typing.Optional[str]
    producer_id: typing.Optional[str]
    # TODO support hash once saved as a column in the artifacts table
    # hash: typing.Optional[str]


class ArtifactMetadata(pydantic.v1.BaseModel):
    key: str
    project: str
    iter: typing.Optional[int]
    tree: typing.Optional[str]
    tag: typing.Optional[str]

    class Config:
        extra = pydantic.v1.Extra.allow


class ArtifactSpec(pydantic.v1.BaseModel):
    src_path: typing.Optional[str]
    target_path: typing.Optional[str]
    viewer: typing.Optional[str]
    inline: typing.Optional[str]
    size: typing.Optional[int]
    db_key: typing.Optional[str]
    extra_data: typing.Optional[dict[str, typing.Any]]
    unpackaging_instructions: typing.Optional[dict[str, typing.Any]]
    parent_uri: typing.Optional[str]

    class Config:
        extra = pydantic.v1.Extra.allow


class Artifact(pydantic.v1.BaseModel):
    kind: str
    metadata: ArtifactMetadata
    spec: ArtifactSpec
    status: ObjectStatus


class ArtifactsDeletionStrategies(mlrun.common.types.StrEnum):
    """Artifacts deletion strategies types."""

    metadata_only = "metadata-only"
    """Only removes the artifact db record, leaving all related artifact data in-place"""

    data_optional = "data-optional"
    """Delete the artifact data of the artifact as a best-effort.
    If artifact data deletion fails still try to delete the artifact db record"""

    data_force = "data-force"
    """Delete the artifact data, and if cannot delete it fail the deletion
    and don’t delete the artifact db record"""



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Optional, Union

import mlrun
from mlrun.datastore.model_provider.model_provider import (
    InvokeResponseFormat,
    ModelProvider,
    UsageResponseKeys,
)

if TYPE_CHECKING:
    from transformers.pipelines.base import Pipeline
    from transformers.pipelines.text_generation import ChatType


class HuggingFaceProvider(ModelProvider):
    """
    HuggingFaceProvider is a wrapper around the Hugging Face Transformers pipeline
    that provides an interface for interacting with a wide range of Hugging Face models.

    It supports synchronous operations, enabling flexible integration into various workflows.

    This class extends the ModelProvider base class and implements Hugging Face-specific
    functionality, including pipeline initialization, default text generation operations,
    and custom operations tailored to the Hugging Face Transformers pipeline API.

    Note: The pipeline object will download the model (if not already cached) and load it
    into memory for inference. Ensure you have the required CPU/GPU and memory to use this operation.
    """

    def __init__(
        self,
        parent,
        schema,
        name,
        endpoint="",
        secrets: Optional[dict] = None,
        default_invoke_kwargs: Optional[dict] = None,
    ):
        endpoint = endpoint or mlrun.mlconf.model_providers.huggingface_default_model
        if schema != "huggingface":
            raise mlrun.errors.MLRunInvalidArgumentError(
                "HuggingFaceProvider supports only 'huggingface' as the provider kind."
            )
        super().__init__(
            parent=parent,
            kind=schema,
            name=name,
            endpoint=endpoint,
            secrets=secrets,
            default_invoke_kwargs=default_invoke_kwargs,
        )
        self.options = self.get_client_options()
        self._expected_operation_type = None
        self._download_model()

    @staticmethod
    def _extract_string_output(response: list[dict]) -> str:
        """
        Extracts the first generated string from Hugging Face pipeline output
        """
        if not isinstance(response, list) or len(response) == 0:
            raise ValueError("Empty or invalid pipeline output")
        if len(response) != 1:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "HuggingFaceProvider: extracting string from response is only supported for single-response outputs"
            )
        return response[0].get("generated_text")

    @classmethod
    def parse_endpoint_and_path(cls, endpoint, subpath) -> (str, str):
        if endpoint and subpath:
            endpoint = endpoint + subpath
            # In HuggingFace, "/" in a model name is part of the name — `subpath` is not used.
            subpath = ""
        return endpoint, subpath

    @property
    def client(self) -> Any:
        """
        Lazily return the HuggingFace-pipeline client.

        If the client has not been initialized yet, it will be created
        by calling `load_client`.
        """
        self.load_client()
        return self._client

    def _download_model(self):
        """
        Pre-downloads model files locally to prevent race conditions in multiprocessing.

        Uses snapshot_download with local_dir_use_symlinks=False to ensure proper
        file copying for safe concurrent access across multiple processes.

        :raises:
            ImportError: If huggingface_hub package is not installed.
        """
        try:
            from huggingface_hub import snapshot_download

            # Download the model and tokenizer files directly to the cache.
            snapshot_download(
                repo_id=self.model,
                local_dir_use_symlinks=False,
                token=self._get_secret_or_env("HF_TOKEN") or None,
            )
        except ImportError as exc:
            raise ImportError("huggingface_hub package is not installed") from exc

    def _response_handler(
        self,
        response: Union[str, list],
        invoke_response_format: InvokeResponseFormat = InvokeResponseFormat.FULL,
        messages: Union[str, list[str], "ChatType", list["ChatType"]] = None,
        **kwargs,
    ) -> Union[str, list, dict[str, Any]]:
        """
        Processes and formats the raw response from the HuggingFace pipeline according to the specified format.

        The response should exclude the user’s input (no repetition in the output).
        This can be accomplished by invoking the pipeline with `return_full_text=False`.

        :param response:                The raw response from the HuggingFace pipeline, typically a list of dictionaries
                                        containing generated text sequences.
        :param invoke_response_format:  Determines how the response should be processed and returned. Options:

                                       - STRING: Return only the main generated content as a string,
                                                 for single-answer responses.
                                       - USAGE: Return a dictionary combining the string response with
                                                token usage statistics:

                                       .. code-block:: json

                                       {
                                           "answer": "<generated_text>",
                                           "usage": {
                                               "prompt_tokens": <int>,
                                               "completion_tokens": <int>,
                                               "total_tokens": <int>
                                           }
                                       }

                                       Note: Token counts are estimated after answer generation and
                                       may differ from the actual tokens generated by the model due to
                                       internal decoding behavior and implementation details.

                                       - FULL: Return the full raw response object.

        :param messages:               The original input messages used for token count estimation in USAGE mode.
                                       Can be a string, list of strings, or chat format messages.
        :param kwargs:                 Additional parameters for response processing.

        :return:                       The processed response in the format specified by `invoke_response_format`.
                                       Can be a string, dictionary, or the original response object.

        :raises MLRunInvalidArgumentError: If extracting the string response fails.
        :raises MLRunRuntimeError: If applying the chat template to the model fails during token usage calculation.
        """
        if InvokeResponseFormat.is_str_response(invoke_response_format.value):
            str_response = self._extract_string_output(response)
            if invoke_response_format == InvokeResponseFormat.STRING:
                return str_response
            if invoke_response_format == InvokeResponseFormat.USAGE:
                tokenizer = self.client.tokenizer
                if not isinstance(messages, str):
                    try:
                        messages = tokenizer.apply_chat_template(
                            messages, tokenize=False, add_generation_prompt=True
                        )
                    except Exception as e:
                        raise mlrun.errors.MLRunRuntimeError(
                            f"Failed to apply chat template using the tokenizer for model '{self.model}'. "
                            "This may indicate that the tokenizer does not support chat formatting, "
                            "or that the input format is invalid. "
                            f"Original error: {e}"
                        )
                prompt_tokens = len(tokenizer.encode(messages))
                completion_tokens = len(tokenizer.encode(str_response))
                total_tokens = prompt_tokens + completion_tokens
                usage = {
                    "prompt_tokens": prompt_tokens,
                    "completion_tokens": completion_tokens,
                    "total_tokens": total_tokens,
                }
                response = {
                    UsageResponseKeys.ANSWER: str_response,
                    UsageResponseKeys.USAGE: usage,
                }
        return response

    def load_client(self) -> None:
        """
        Initializes the Hugging Face pipeline using the provided options.

        This method imports the `pipeline` function from the `transformers` package,
        creates a pipeline instance with the specified task and model (from `self.options`),
        and assigns it to `self._client`.

        Note: Hugging Face pipelines are synchronous and do not support async invocation.

        :raises:
            ImportError: If the `transformers` package is not installed.
        """
        if self._client:
            return
        try:
            from transformers import pipeline, AutoModelForCausalLM  # noqa
            from transformers import AutoTokenizer  # noqa
            from transformers.pipelines.base import Pipeline  # noqa

            self.options["model_kwargs"] = self.options.get("model_kwargs", {})
            self.options["model_kwargs"]["local_files_only"] = True
            self._client = pipeline(model=self.model, **self.options)
            self._expected_operation_type = Pipeline
        except ImportError as exc:
            raise ImportError("transformers package is not installed") from exc

    def get_client_options(self):
        res = dict(
            task=self._get_secret_or_env("HF_TASK") or "text-generation",
            token=self._get_secret_or_env("HF_TOKEN"),
            device=self._get_secret_or_env("HF_DEVICE"),
            device_map=self._get_secret_or_env("HF_DEVICE_MAP"),
            trust_remote_code=self._get_secret_or_env("HF_TRUST_REMOTE_CODE"),
            model_kwargs=self._get_secret_or_env("HF_MODEL_KWARGS"),
        )
        return self._sanitize_options(res)

    def custom_invoke(
        self, operation: Optional["Pipeline"] = None, **invoke_kwargs
    ) -> Union[list, dict, Any]:
        """
        Invokes a HuggingFace pipeline operation with the given keyword arguments.

        This method provides flexibility to use a custom pipeline object for specific tasks
        (e.g., image classification, sentiment analysis).

        The operation must be a Pipeline object from the transformers library that accepts keyword arguments.

        Example:
            ```python
            from transformers import pipeline
            from PIL import Image

            # Using custom pipeline for image classification
            image = Image.open(image_path)
            pipeline_object = pipeline("image-classification", model="microsoft/resnet-50")
            result = hf_provider.custom_invoke(
                pipeline_object,
                inputs=image,
            )
            ```

        :param operation:      A Pipeline object from the transformers library.
                               If not provided, defaults to the provider's configured pipeline.
        :param invoke_kwargs:  Keyword arguments to pass to the pipeline operation.
                               These are merged with `default_invoke_kwargs` and may include
                               parameters such as `inputs`, `max_length`, `temperature`, or task-specific options.

        :return:               The full response returned by the pipeline operation.
                               Format depends on the pipeline task (list for text generation,
                               dict for classification, etc.).

        :raises MLRunInvalidArgumentError: If the operation is not a valid Pipeline object.

        """
        invoke_kwargs = self.get_invoke_kwargs(invoke_kwargs)
        if operation:
            if not isinstance(operation, self._expected_operation_type):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Huggingface operation must inherit" " from 'Pipeline' object"
                )
            return operation(**invoke_kwargs)
        else:
            return self.client(**invoke_kwargs)

    def invoke(
        self,
        messages: Union[str, list[str], "ChatType", list["ChatType"]],
        invoke_response_format: InvokeResponseFormat = InvokeResponseFormat.FULL,
        **invoke_kwargs,
    ) -> Union[str, list, dict[str, Any]]:
        """
        HuggingFace-specific implementation of model invocation using the synchronous pipeline client.
        Invokes a HuggingFace model operation for text generation tasks.

        Note: Ensure your environment has sufficient computational resources (CPU/GPU and memory) to run the model.

        :param messages:
            Input for the text generation model. Can be provided in multiple formats:

            - A single string: Direct text input for generation
            - A list of strings: Multiple text inputs for batch processing
            - Chat format: A list of dictionaries with "role" and "content" keys:

            .. code-block:: json

                [
                    {"role": "system", "content": "You are a helpful assistant."},
                    {"role": "user", "content": "What is the capital of France?"}
                ]

        :param invoke_response_format: InvokeResponseFormat
            Specifies the format of the returned response. Options:

            - "string": Returns only the generated text content, extracted from a single response.
            - "usage":  Combines the generated text with metadata (e.g., token usage), returning a dictionary:

            .. code-block:: json
                {
                    "answer": "<generated_text>",
                    "usage": {
                        "prompt_tokens": <int>,
                        "completion_tokens": <int>,
                        "total_tokens": <int>
                    }
                }

            Note: For usage mode, the model tokenizer should support apply_chat_template.

            - "full":   Returns the raw response object from the HuggingFace model,
                        typically a list of generated sequences (dictionaries).
                        This format does not include token usage statistics.

        :param invoke_kwargs:
            Additional keyword arguments passed to the HuggingFace pipeline.

        :return:
            A string, dictionary, or list of model outputs, depending on `invoke_response_format`.

        :raises MLRunInvalidArgumentError:
            If the pipeline task is not "text-generation" or if the response contains multiple outputs when extracting
            string content.
        :raises MLRunRuntimeError:
            If using "usage" response mode and the model tokenizer does not support chat template formatting.
        """
        if self.client.task != "text-generation":
            raise mlrun.errors.MLRunInvalidArgumentError(
                "HuggingFaceProvider.invoke supports text-generation task only"
            )
        if InvokeResponseFormat.is_str_response(invoke_response_format.value):
            invoke_kwargs["return_full_text"] = False
        response = self.custom_invoke(text_inputs=messages, **invoke_kwargs)
        response = self._response_handler(
            messages=messages,
            response=response,
            invoke_response_format=invoke_response_format,
        )
        return response



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mlrun
import mlrun.artifacts
from mlrun.config import config
from mlrun.utils.helpers import parse_artifact_uri

from ..common.helpers import parse_versioned_object_uri
from ..platforms.iguazio import parse_path
from ..utils import DB_SCHEMA, StorePrefix
from .targets import get_online_target


def is_store_uri(url):
    """detect if the uri starts with the store schema prefix"""
    if not url:
        return False
    return url.startswith(DB_SCHEMA + "://")


def parse_store_uri(url):
    """parse a store uri and return kind + uri suffix"""
    if not is_store_uri(url):
        return None, ""
    uri = url[len(DB_SCHEMA) + len("://") :].strip("/")
    split = uri.split("/", 1)
    if len(split) == 0:
        raise ValueError(f"url {url} has no path")
    if split and StorePrefix.is_prefix(split[0]):
        return split[0], split[1]
    return StorePrefix.Artifact, uri


def get_store_uri(kind, uri):
    """return uri from store kind and suffix"""
    return f"{DB_SCHEMA}://{kind}/{uri}"


class ResourceCache:
    """Resource cache for real-time pipeline/serving and storey
    this cache is basic and doesnt have sync or ttl logic
    """

    def __init__(self):
        self._tabels = {}
        self._resources = {}

    def cache_table(self, uri, value, is_default=False):
        """Cache storey Table objects"""
        self._tabels[uri] = value
        if is_default:
            self._tabels["."] = value

    def get_table(self, uri):
        """get storey Table object by uri"""
        try:
            from storey import Driver, Table, V3ioDriver
        except ImportError:
            raise ImportError("storey package is not installed, use pip install storey")
        if uri in self._tabels:
            return self._tabels[uri]
        if uri in [".", ""] or uri.startswith("$"):  # $.. indicates in-mem table
            self._tabels[uri] = Table("", Driver())
            return self._tabels[uri]

        if uri.startswith("v3io://") or uri.startswith("v3ios://"):
            endpoint, uri = parse_path(uri)
            self._tabels[uri] = Table(
                uri,
                V3ioDriver(webapi=endpoint or mlrun.mlconf.v3io_api),
                flush_interval_secs=mlrun.mlconf.feature_store.flush_interval,
            )
            return self._tabels[uri]

        if uri.startswith("redis://") or uri.startswith("rediss://"):
            from storey.redis_driver import RedisDriver

            endpoint, uri = parse_path(uri)
            endpoint = endpoint or mlrun.mlconf.redis.url
            self._tabels[uri] = Table(
                uri,
                RedisDriver(redis_url=endpoint, key_prefix="/"),
                flush_interval_secs=mlrun.mlconf.feature_store.flush_interval,
            )
            return self._tabels[uri]

        if is_store_uri(uri):
            resource = get_store_resource(uri)
            if resource.kind in [
                mlrun.common.schemas.ObjectKind.feature_set.value,
                mlrun.common.schemas.ObjectKind.feature_vector.value,
            ]:
                target = get_online_target(resource)
                if not target:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"resource {uri} does not have an online data target"
                    )
                self._tabels[uri] = target.get_table_object()
                return self._tabels[uri]

        raise mlrun.errors.MLRunInvalidArgumentError(f"table {uri} not found in cache")

    def cache_resource(self, uri, value, default=False):
        """cache store resource (artifact/feature-set/feature-vector)"""
        self._resources[uri] = value
        if default:
            self._resources["."] = value

    def get_resource(self, uri):
        """get resource from cache by uri"""
        return self._resources[uri]

    def resource_getter(self, db=None, secrets=None):
        """wraps get_store_resource with a simple object cache"""

        def _get_store_resource(uri, use_cache=True):
            """get mlrun store resource object
            :param use_cache: indicate if we read from local cache or from DB
            """
            if (uri == "." or use_cache) and uri in self._resources:
                return self._resources[uri]
            resource = get_store_resource(uri, db, secrets=secrets)
            if use_cache:
                self._resources[uri] = resource
            return resource

        return _get_store_resource


def get_store_resource(
    uri, db=None, secrets=None, project=None, data_store_secrets=None
):
    """get store resource object by uri"""

    db = db or mlrun.get_run_db(secrets=secrets)
    kind, uri = parse_store_uri(uri)
    if not kind:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"Cannot get store resource from invalid URI: {uri}"
        )
    elif kind == StorePrefix.FeatureSet:
        project, name, tag, uid = parse_versioned_object_uri(
            uri, project or config.active_project
        )
        return db.get_feature_set(name, project, tag, uid)

    elif kind == StorePrefix.FeatureVector:
        project, name, tag, uid = parse_versioned_object_uri(
            uri, project or config.active_project
        )
        return db.get_feature_vector(name, project, tag, uid)

    elif StorePrefix.is_artifact(kind):
        project, key, iteration, tag, tree, uid = parse_artifact_uri(
            uri, project or config.active_project
        )
        resource = db.read_artifact(
            key,
            project=project,
            tag=tag,
            iter=iteration,
            tree=tree,
            uid=uid,
        )
        if resource.get("kind", "") == "link":
            # todo: support other link types (not just iter, move this to the db/api layer
            link_iteration = resource["spec"].get("link_iteration", 0)

            resource = db.read_artifact(
                key,
                tag=tag,
                iter=link_iteration,
                project=project,
            )
        if resource:
            return mlrun.artifacts.dict_to_artifact(resource)

    else:
        stores = mlrun.store_manager.set(secrets, db=db)
        return stores.object(url=uri, secrets=data_store_secrets)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC

from ...sklearn import SKLearnMLRunInterface


class LGBMModelMLRunInterface(SKLearnMLRunInterface, ABC):
    """
    Interface for adding MLRun features for LightGBM models (SciKit-Learn API).
    """

    # TODO: Should be changed from SKLearn's interface to its own, it has the same `params` and callbacks passed to
    #       `train`.
    # TODO: Add to `apply_mlrun` a "use_dask": bool = None argument. A boolean value that will replace the object of a
    #       SciKit-Learn API `LGBMModel` to its Dask version (`LGBMClassifier` to `DaskLGBMClassifier`). None will look
    #       for dask parameters in the given context and turn on and off accordingly.
    # MLRun's context default name:
    DEFAULT_CONTEXT_NAME = "mlrun-lgbm"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import threading
import time
import traceback
from typing import Optional

import mlrun.artifacts
import mlrun.common.model_monitoring.helpers
import mlrun.common.schemas.model_monitoring
import mlrun.model_monitoring
from mlrun.utils import logger, now_date

from ..common.model_monitoring.helpers import (
    get_model_endpoints_creation_task_status,
)
from .utils import StepToDict, _extract_input_data, _update_result_body


class V2ModelServer(StepToDict):
    def __init__(
        self,
        context=None,
        name: Optional[str] = None,
        model_path: Optional[str] = None,
        model=None,
        protocol=None,
        input_path: Optional[str] = None,
        result_path: Optional[str] = None,
        shard_by_endpoint: Optional[bool] = None,
        **kwargs,
    ):
        """base model serving class (v2), using similar API to KFServing v2 and Triton

        The class is initialized automatically by the model server and can run locally
        as part of a nuclio serverless function, or as part of a real-time pipeline
        default model url is: /v2/models/<model>[/versions/<ver>]/operation

        You need to implement two mandatory methods:
          load()     - download the model file(s) and load the model into memory
          predict()  - accept request payload and return prediction/inference results

        you can override additional methods : preprocess, validate, postprocess, explain
        you can add custom api endpoint by adding method op_xx(event), will be invoked by
        calling the <model-url>/xx (operation = xx)

        model server classes are subclassed (subclass implements the `load()` and `predict()` methods)
        the subclass can be added to a serving graph or to a model router

        defining a sub class::

            class MyClass(V2ModelServer):
                def load(self):
                    # load and initialize the model and/or other elements
                    model_file, extra_data = self.get_model(suffix=".pkl")
                    self.model = load(open(model_file, "rb"))

                def predict(self, request):
                    events = np.array(request["inputs"])
                    dmatrix = xgb.DMatrix(events)
                    result: xgb.DMatrix = self.model.predict(dmatrix)
                    return {"outputs": result.tolist()}

        usage example::

            # adding a model to a serving graph using the subclass MyClass
            # MyClass will be initialized with the name "my", the model_path, and an arg called my_param
            graph = fn.set_topology("router")
            fn.add_model("my", class_name="MyClass", model_path="<model-uri>>", my_param=5)

        :param context:    for internal use (passed in init)
        :param name:       step name
        :param model_path: model file/dir or artifact path
        :param model:      model object (for local testing)
        :param protocol:   serving API protocol (default "v2")
        :param input_path:    when specified selects the key/path in the event to use as body
                              this require that the event body will behave like a dict, example:
                              event: {"data": {"a": 5, "b": 7}}, input_path="data.b" means request body will be 7
        :param result_path:   selects the key/path in the event to write the results to
                              this require that the event body will behave like a dict, example:
                              event: {"x": 5} , result_path="resp" means the returned response will be written
                              to event["y"] resulting in {"x": 5, "resp": <result>}
        :param shard_by_endpoint: whether to use the endpoint as the partition/sharding key when writing to model
                                  monitoring stream. Defaults to True.
        :param kwargs:     extra arguments (can be accessed using self.get_param(key))
        """
        self.name = name
        self.context = context
        self.ready = False
        self.error = ""
        self.protocol = protocol or "v2"
        self.model_path = model_path
        self.model_spec: Optional[mlrun.artifacts.ModelArtifact] = None
        self._input_path = input_path
        self._result_path = result_path
        self._kwargs = kwargs  # for to_dict()
        self._params = kwargs
        self.metrics = {}
        self.labels = {}
        self.model = None
        if model:
            self.model = model
            self.ready = True
        self.model_endpoint_uid = kwargs.get("model_endpoint_uid", None)
        self.shard_by_endpoint = shard_by_endpoint
        self._model_logger = None
        self.initialized = False
        self.output_schema = kwargs.get("outputs", [])

    def _load_and_update_state(self):
        try:
            self.load()
        except Exception as exc:
            self.error = exc
            self.context.logger.error(traceback.format_exc())
            raise RuntimeError(f"failed to load model {self.name}") from exc
        self.ready = True
        self.context.logger.info(f"model {self.name} was loaded")

    def post_init(self, mode="sync", **kwargs):
        """sync/async model loading, for internal use"""
        if not self.ready:
            if mode == "async":
                t = threading.Thread(target=self._load_and_update_state)
                t.start()
                self.context.logger.info(f"started async model loading for {self.name}")
            else:
                self._load_and_update_state()

        if self.ready and not self.context.is_mock and not self.model_spec:
            self.get_model()

        if self.model_spec:
            self.output_schema = self.output_schema or [
                feature.name for feature in self.model_spec.outputs
            ]

        if (
            kwargs.get("endpoint_type", mlrun.common.schemas.EndpointType.LEAF_EP)
            == mlrun.common.schemas.EndpointType.NODE_EP
        ):
            self._initialize_model_logger()

    def _lazy_init(self, event):
        if event and isinstance(event, dict) and not self.initialized:
            background_task_state = event.get("background_task_state", None)
            if (
                background_task_state
                == mlrun.common.schemas.BackgroundTaskState.succeeded
            ):
                self._model_logger = (
                    _ModelLogPusher(self, self.context)
                    if self.context
                    and self.context.stream.enabled
                    and self.model_endpoint_uid
                    else None
                )
                self.initialized = True

    def get_param(self, key: str, default=None):
        """get param by key (specified in the model or the function)"""
        if key in self._params:
            return self._params.get(key)
        return self.context.get_param(key, default=default)

    def set_metric(self, name: str, value):
        """set real time metric (for model monitoring)"""
        self.metrics[name] = value

    def get_model(self, suffix="") -> (str, dict):
        """get the model file(s) and metadata from model store

        the method returns a path to the model file and the extra data (dict of dataitem objects)
        it also loads the model metadata into the self.model_spec attribute, allowing direct access
        to all the model metadata attributes.

        get_model is usually used in the model .load() method to init the model
        Examples
        --------
        ::

            def load(self):
                model_file, extra_data = self.get_model(suffix=".pkl")
                self.model = load(open(model_file, "rb"))
                categories = extra_data["categories"].as_df()

        Parameters
        ----------
        suffix : str
            optional, model file suffix (when the model_path is a directory)

        Returns
        -------
        str
            (local) model file
        dict
            extra dataitems dictionary

        """
        if self.model_path:
            model_file, self.model_spec, extra_dataitems = mlrun.artifacts.get_model(
                self.model_path, suffix
            )
            if self.model_spec and self.model_spec.parameters:
                for key, value in self.model_spec.parameters.items():
                    self._params[key] = value
            return model_file, extra_dataitems
        return None, None

    def load(self):
        """model loading function, see also .get_model() method"""
        if not self.ready and not self.model:
            raise ValueError("please specify a load method or a model object")

    def _check_readiness(self, event):
        if self.ready:
            return
        if not event.trigger or event.trigger.kind in ["http", ""]:
            raise RuntimeError(f"model {self.name} is not ready yet")
        self.context.logger.info(f"waiting for model {self.name} to load")
        for i in range(50):  # wait up to 4.5 minutes
            time.sleep(5)
            if self.ready:
                return
        raise RuntimeError(f"model {self.name} is not ready {self.error}")

    def _pre_event_processing_actions(self, event, event_body, op):
        self._check_readiness(event)
        if "_dict" in op:
            event_body = self._inputs_to_list(event_body)
        request = self.preprocess(event_body, op)
        return self.validate(request, op)

    def do_event(self, event, *args, **kwargs):
        """main model event handler method"""
        if not self.initialized:
            self._lazy_init(event.body)
        start = now_date()
        original_body = event.body
        event_body = _extract_input_data(self._input_path, event.body)
        event_id = event.id
        op = event.path.strip("/")

        partition_key = (
            self.model_endpoint_uid if self.shard_by_endpoint is not False else None
        )

        if event_body and isinstance(event_body, dict):
            op = op or event_body.get("operation")
            event_id = event_body.get("id", event_id)
        if not op and event.method != "GET":
            op = "infer"

        if (
            op == "predict"
            or op == "infer"
            or op == "infer_dict"
            or op == "predict_dict"
        ):
            # predict operation
            request = self._pre_event_processing_actions(event, event_body, op)
            try:
                outputs = self.predict(request)
            except Exception as exc:
                request["id"] = event_id
                if self._model_logger:
                    self._model_logger.push(
                        start,
                        request,
                        op=op,
                        error=exc,
                        partition_key=partition_key,
                    )
                raise exc

            response = {
                "id": event_id,
                "model_name": self.name.split(":")[0],
                "outputs": outputs,
                "timestamp": start.isoformat(sep=" ", timespec="microseconds"),
            }
            if self.model_endpoint_uid:
                response["model_endpoint_uid"] = self.model_endpoint_uid
        elif op == "ready" and event.method == "GET":
            # get model health operation
            setattr(event, "terminated", True)
            if self.ready:
                # Generate a response, confirming that the model is ready
                event.body = self.context.Response(
                    status_code=200,
                    body=bytes(
                        f"Model {self.name} is ready (event_id = {event_id})",
                        encoding="utf-8",
                    ),
                )

            else:
                event.body = self.context.Response(
                    status_code=408, body=b"model not ready"
                )

            return event

        elif op == "" and event.method == "GET":
            # get model metadata operation
            setattr(event, "terminated", True)
            event_body = {
                "name": self.name.split(":")[0],
                "model_endpoint_uid": self.model_endpoint_uid or "",
                "inputs": [],
                "outputs": [],
            }
            if self.model_spec:
                event_body["inputs"] = self.model_spec.inputs.to_dict()
                event_body["outputs"] = self.model_spec.outputs.to_dict()
            event.body = _update_result_body(
                self._result_path, original_body, event_body
            )
            return event

        elif op == "explain":
            # explain operation
            request = self._pre_event_processing_actions(event, event_body, op)
            try:
                outputs = self.explain(request)
            except Exception as exc:
                request["id"] = event_id
                if self._model_logger:
                    self._model_logger.push(
                        start,
                        request,
                        op=op,
                        error=exc,
                        partition_key=partition_key,
                    )
                raise exc

            response = {
                "id": event_id,
                "model_name": self.name,
                "outputs": outputs,
            }
            if self.model_endpoint_uid:
                response["model_endpoint_uid"] = self.model_endpoint_uid

        elif hasattr(self, "op_" + op):
            # custom operation (child methods starting with "op_")
            response = getattr(self, "op_" + op)(event)
            event.body = _update_result_body(self._result_path, original_body, response)
            return event

        else:
            raise ValueError(f"illegal model operation {op}, method={event.method}")

        response = self.postprocess(response)
        if self._model_logger:
            inputs, outputs = self.logged_results(request, response, op)
            if inputs is None and outputs is None:
                self._model_logger.push(
                    start, request, response, op, partition_key=partition_key
                )
            else:
                track_request = {"id": event_id, "inputs": inputs or []}
                track_response = {"outputs": outputs or []}
                # TODO : check dict/list
                self._model_logger.push(
                    start,
                    track_request,
                    track_response,
                    op,
                    partition_key=partition_key,
                )
        event.body = _update_result_body(self._result_path, original_body, response)
        return event

    def logged_results(self, request: dict, response: dict, op: str):
        """Hook for controlling which results are tracked by the model monitoring

        This hook allows controlling which input/output data is logged by the model monitoring.
        It allows filtering out columns or adding custom values, and can also be used to monitor derived metrics,
        for example in image classification to calculate and track the RGB values vs the image bitmap.

        The request ["inputs"] holds a list of input values/arrays, the response ["outputs"] holds a list of
        corresponding output values/arrays (the schema of the input/output fields is stored in the model object).
        This method should return lists of alternative inputs and outputs which will be monitored.

        :param request:   predict/explain request, see model serving docs for details
        :param response:  result from the model predict/explain (after postprocess())
        :param op:        operation (predict/infer or explain)
        :returns: the input and output lists to track
        """
        return None, None

    def validate(self, request, operation):
        """validate the event body (after preprocess)"""
        if self.protocol == "v2":
            if "inputs" not in request:
                raise Exception('Expected key "inputs" in request body')

            if not isinstance(request["inputs"], list):
                raise Exception('Expected "inputs" to be a list')

        return request

    def preprocess(self, request: dict, operation) -> dict:
        """preprocess the event body before validate and action"""
        return request

    def postprocess(self, request: dict) -> dict:
        """postprocess, before returning response"""
        return request

    def predict(self, request: dict) -> list:
        """model prediction operation

        :return: list with the model prediction results (can be multi-port) or list of lists for multiple predictions
        """
        raise NotImplementedError()

    def explain(self, request: dict) -> dict:
        """model explain operation"""
        raise NotImplementedError()

    def _inputs_to_list(self, request: dict) -> dict:
        """
        Convert the inputs from list of dictionary / dictionary to list of lists / list
        where the internal list order is according to the ArtifactModel inputs.

        :param request: event
        :return: event body converting the inputs to be list of lists
        """
        if self.model_spec and self.model_spec.inputs:
            input_order = [feature.name for feature in self.model_spec.inputs]
        else:
            raise mlrun.MLRunInvalidArgumentError(
                "In order to use predict_dict or infer_dict operation you have to provide `model_path` "
                "to the model server and to load it by `load()` function"
            )
        inputs = request.get("inputs")
        try:
            if isinstance(inputs, list) and all(
                isinstance(item, dict) for item in inputs
            ):
                new_inputs = [
                    [input_dict[key] for key in input_order] for input_dict in inputs
                ]
            elif isinstance(inputs, dict):
                new_inputs = [inputs[key] for key in input_order]
            else:
                raise mlrun.MLRunInvalidArgumentError(
                    "When using predict_dict or infer_dict operation the inputs must be "
                    "of type `list[dict]` or `dict`"
                )
        except KeyError:
            raise mlrun.MLRunInvalidArgumentError(
                f"Input dictionary don't contain all the necessary input keys : {input_order}"
            )
        request["inputs"] = new_inputs
        return request

    def _initialize_model_logger(self):
        server: mlrun.serving.GraphServer = getattr(
            self.context, "_server", None
        ) or getattr(self.context, "server", None)
        if not self.context.is_mock or self.context.monitoring_mock:
            if server.model_endpoint_creation_task_name:
                background_task_state, _, _ = get_model_endpoints_creation_task_status(
                    server
                )
                if (
                    background_task_state
                    in mlrun.common.schemas.BackgroundTaskState.terminal_states()
                ):
                    logger.debug(
                        f"Model endpoint creation task completed with state {background_task_state}"
                    )
                    if (
                        background_task_state
                        == mlrun.common.schemas.BackgroundTaskState.succeeded
                    ):
                        self._model_logger = (
                            _ModelLogPusher(self, self.context)
                            if self.context
                            and self.context.stream.enabled
                            and self.model_endpoint_uid
                            else None
                        )
                        self.initialized = True

                else:  # in progress
                    logger.debug(
                        f"Model endpoint creation task is still in progress with the current state: "
                        f"{background_task_state}.",
                        name=self.name,
                    )
            else:
                logger.error(
                    "Model endpoint creation task name not provided. This function is not being monitored.",
                )


class _ModelLogPusher:
    def __init__(self, model: V2ModelServer, context, output_stream=None):
        self.model = model
        self.verbose = context.verbose
        self.hostname = context.stream.hostname
        self.function_uri = context.stream.function_uri
        self.sampling_percentage = float(context.get_param("sampling_percentage", 100))
        self.output_stream = output_stream or context.stream.output_stream
        self._worker = context.worker_id

    def base_data(self):
        base_data = {
            "class": self.model.__class__.__name__,
            "worker": self._worker,
            "model": self.model.name,
            "host": self.hostname,
            "function_uri": self.function_uri,
            "endpoint_id": self.model.model_endpoint_uid,
            "sampling_percentage": self.sampling_percentage,
        }
        if getattr(self.model, "labels", None):
            base_data["labels"] = self.model.labels
        return base_data

    def push(self, start, request, resp=None, op=None, error=None, partition_key=None):
        start_str = start.isoformat(sep=" ", timespec="microseconds")
        if error:
            data = self.base_data()
            data["request"] = request
            data["op"] = op
            data["when"] = start_str
            message = str(error)
            if self.verbose:
                message = f"{message}\n{traceback.format_exc()}"
            data["error"] = message
            self.output_stream.push([data], partition_key=partition_key)
            return

        if self.output_stream:
            # Ensure that the inputs are a list of lists
            request["inputs"] = (
                request["inputs"]
                if not any(not isinstance(req, list) for req in request["inputs"])
                else [request["inputs"]]
            )
            microsec = (now_date() - start).microseconds

            if self.sampling_percentage != 100:
                # Randomly select a subset of the requests based on the percentage
                num_of_inputs = len(request["inputs"])
                sampled_requests_indices = self._pick_random_requests(
                    num_of_inputs, self.sampling_percentage
                )
                if not sampled_requests_indices:
                    # No events were selected for sampling
                    return

                request["inputs"] = [
                    request["inputs"][i] for i in sampled_requests_indices
                ]

                if resp and "outputs" in resp and isinstance(resp["outputs"], list):
                    resp["outputs"] = [
                        resp["outputs"][i] for i in sampled_requests_indices
                    ]
                if self.model.output_schema and len(self.model.output_schema) != len(
                    resp["outputs"][0]
                ):
                    logger.info(
                        "The number of outputs returned by the model does not match the number of outputs "
                        "specified in the model endpoint.",
                        model_endpoint=self.model.name,
                        model_endpoint_id=self.model.model_endpoint_uid,
                        output_len=len(resp["outputs"][0]),
                        schema_len=len(self.model.output_schema),
                    )

            data = self.base_data()
            data["request"] = request
            data["op"] = op
            data["resp"] = resp
            data["when"] = start_str
            data["microsec"] = microsec
            if getattr(self.model, "metrics", None):
                data["metrics"] = self.model.metrics
            data["effective_sample_count"] = len(request["inputs"])
            self.output_stream.push([data], partition_key=partition_key)

    @staticmethod
    def _pick_random_requests(num_of_reqs: int, percentage: float) -> list[int]:
        """
        Randomly selects indices of requests to sample based on the given percentage

        :param num_of_reqs: Number of requests to select from
        :param percentage: Sample percentage for each request
        :return: A list containing the indices of the selected requests
        """

        return [
            req for req in range(num_of_reqs) if random.random() < (percentage / 100)
        ]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import storey


class CacheEntry:
    def __init__(self, sequence, event):
        self.arrived = 1
        self.sequence = sequence
        self.events = [event]
        self.is_late = False

    def add_event(self, context, event, merge_key):
        if self.is_late:
            context.logger.warning(f"event id {self} arrived late")
            return None

        if context.verbose:
            context.logger.info(f"event id {merge_key} part {self.arrived} arrived")
        self.arrived += 1
        self.events.append(event)


class Merge(storey.Flow):
    def __init__(
        self,
        full_event: Optional[bool] = None,
        key_path: Optional[str] = None,
        max_behind: Optional[int] = None,
        expected_num_events: Optional[int] = None,
        **kwargs,
    ):
        """Merge multiple events based on event id or provided key path

        Users can subclass and overwrite the `get_join_key()` and `merge_function()` with custom logic

        :param key_path:   path to the event join key e.g. 'event["doc_id"]', default is the unique event id
        :param max_behind: max queue size to hold unmerged events,
                           oldest events will be dropped is queued longer (default=64)
        :param expected_num_events:  manually set the expected number of events per key
                                     (keep blank to auto detect from the graph)
        :param full_event: this step accepts the full Event object (body + metadata), not just body
        :param kwargs:     reserved for system use
        """
        self.key_path = key_path
        self.max_behind = max_behind
        self.expected_num_events = expected_num_events

        # use event.id (require full event) by default
        if full_event is None and not key_path:
            full_event = True
        self._graph_step = kwargs.pop("graph_step", None)
        super().__init__(full_event=full_event, **kwargs)

        self._uplinks = None
        self._cache = {}
        self._sequence = 0
        self._get_join_key = None
        self._queue_len = max_behind or 64  # default queue is 64 entries
        self._keys_queue = []

    def post_init(self, mode="sync", **kwargs):
        # auto detect number of uplinks or use user specified value
        self._uplinks = self.expected_num_events or (
            len(self._graph_step.after) if self._graph_step else 0
        )

        # function to extract the join key from the event
        key_path = self.key_path or "event.id"
        self._get_join_key = eval("lambda event: " + key_path, {}, {})
        self._keys_queue = [None] * self._queue_len

    def get_join_key(self, event):
        """function extract the join key from the event, can be overwritten by sub class"""
        return self._get_join_key(event)

    async def _do(self, event):
        if event is storey.dtypes._termination_obj:
            return await self._do_downstream(storey.dtypes._termination_obj)
        else:
            element = self._get_event_or_body(event)
            fn_result = self._merge_events(element)
            if fn_result:
                mapped_event = self._user_fn_output_to_event(event, fn_result)
                await self._do_downstream(mapped_event)

    def _merge_events(self, event):
        # skip if only one uplink
        if self._uplinks <= 1:
            return event

        merge_key = self.get_join_key(event)
        if merge_key in self._cache:
            # old events with that key already exist (cached)
            entry: CacheEntry = self._cache[merge_key]
            entry.add_event(self.context, event, merge_key)

            if entry.arrived == self._uplinks:
                # expected number of events arrived, can merge them
                if self.context.verbose:
                    self.context.logger.info(
                        f"event {merge_key}, all {entry.arrived} parts arrived"
                    )
                del self._cache[merge_key]
                self._keys_queue[entry.sequence % self._queue_len] = None
                return self.merge_function(event, entry.events)
        else:
            # first time the key arrives
            self._cache[merge_key] = CacheEntry(self._sequence, event)
            queue_index = self._sequence % self._queue_len
            old_key = self._keys_queue[queue_index]

            if self.context.verbose:
                self.context.logger.info(f"new event id {merge_key} arrived")

            if old_key is not None:
                # reached max queued entries, need to drop the old event
                message = (
                    f"missing parts for event key {old_key} after a long wait, dropping"
                )
                self.context.logger.warning(message)
                if self._full_event:
                    self.context.push_error(
                        self._cache[old_key].events[0], f"{message}", source=self.name
                    )
                self._cache[old_key].is_late = True

            self._keys_queue[queue_index] = merge_key
            self._sequence += 1
            return None

    def merge_function(self, last_event, events):
        """logic to merge all gathered events to a result event, can be overwritten by sub class

        :param last_event: latest event
        :param events: all gathered events for the same key (including latest)
        :return: event
        """
        if self._full_event:
            last_event.body = [event.body for event in events]
            return last_event
        else:
            return events



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import logging
import typing
from copy import copy
from datetime import datetime
from enum import Enum
from typing import Union

import pandas as pd

import mlrun

from ..config import config as mlconf
from ..datastore import get_store_uri
from ..datastore.targets import BaseStoreTarget, get_offline_target
from ..feature_store.common import (
    get_feature_set_by_uri,
    parse_feature_string,
    parse_project_name_from_feature_string,
    verify_feature_vector_permissions,
)
from ..feature_store.feature_set import FeatureSet
from ..features import Entity, Feature
from ..model import (
    DataSource,
    DataTarget,
    DataTargetBase,
    ModelObj,
    ObjectDict,
    ObjectList,
    VersionedObjMetadata,
)
from ..runtimes.function_reference import FunctionReference
from ..serving.states import RootFlowStep
from ..utils import StorePrefix
from .common import RunConfig
from .feature_vector_utils import JoinGraph, OnlineVectorService
from .retrieval import get_merger, run_merge_job


def _features_to_vector_and_check_permissions(features: "FeatureVector", update_stats):
    vector = features
    if not vector.metadata.name:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "feature vector name must be specified"
        )
    verify_feature_vector_permissions(
        vector, mlrun.common.schemas.AuthorizationAction.update
    )

    vector.save()
    return vector


class FeatureVectorSpec(ModelObj):
    def __init__(
        self,
        features=None,
        description=None,
        entity_source=None,
        entity_fields=None,
        timestamp_field=None,
        graph=None,
        label_feature=None,
        with_indexes=None,
        function=None,
        analysis=None,
        relations=None,
        join_graph=None,
    ):
        self._graph: RootFlowStep = None
        self._entity_fields: ObjectList = None
        self._entity_source: DataSource = None
        self._function: FunctionReference = None
        self._relations: dict[str, ObjectDict] = None
        self._join_graph: JoinGraph = None

        self.description = description
        self.features: list[str] = features or []
        self.entity_source = entity_source
        self.entity_fields = entity_fields or []
        self.graph = graph
        self.join_graph = join_graph
        self.relations: dict[str, dict[str, Union[Entity, str]]] = relations or {}
        self.timestamp_field = timestamp_field
        self.label_feature = label_feature
        self.with_indexes = with_indexes
        self.function = function
        self.analysis = analysis or {}

    @property
    def entity_source(self) -> DataSource:
        """data source used as entity source (events/keys need to be enriched)"""
        return self._entity_source

    @entity_source.setter
    def entity_source(self, source: DataSource):
        self._entity_source = self._verify_dict(source, "entity_source", DataSource)

    @property
    def entity_fields(self) -> list[Feature]:
        """the schema/metadata for the entity source fields"""
        return self._entity_fields

    @entity_fields.setter
    def entity_fields(self, entity_fields: list[Feature]):
        self._entity_fields = ObjectList.from_list(Feature, entity_fields)

    @property
    def graph(self) -> RootFlowStep:
        """feature vector transformation graph/DAG"""
        return self._graph

    @graph.setter
    def graph(self, graph):
        self._graph = self._verify_dict(graph, "graph", RootFlowStep)
        self._graph.engine = "async"

    @property
    def function(self) -> FunctionReference:
        """reference to template graph processing function"""
        return self._function

    @function.setter
    def function(self, function):
        self._function = self._verify_dict(function, "function", FunctionReference)

    @property
    def relations(self) -> dict[str, ObjectDict]:
        """feature set relations dict"""
        return self._relations

    @relations.setter
    def relations(self, relations: dict[str, dict[str, Union[Entity, str]]]):
        temp_relations = {}
        for fs_name, relation in relations.items():
            for col, ent in relation.items():
                if isinstance(ent, str):
                    relation[col] = Entity(ent)
            temp_relations[fs_name] = ObjectDict.from_dict(
                {"entity": Entity}, relation, "entity"
            )
        self._relations = ObjectDict.from_dict(
            {"object_dict": ObjectDict}, temp_relations, "object_dict"
        )

    @property
    def join_graph(self):
        return self._join_graph

    @join_graph.setter
    def join_graph(self, join_graph):
        if join_graph is not None:
            self._join_graph = self._verify_dict(join_graph, "join_graph", JoinGraph)
        else:
            self._join_graph = None


class FeatureVectorStatus(ModelObj):
    def __init__(
        self,
        state=None,
        targets=None,
        features=None,
        label_column=None,
        stats=None,
        preview=None,
        run_uri=None,
        index_keys=None,
        timestamp_key=None,
    ):
        self._targets: ObjectList = None
        self._features: ObjectList = None

        self.state = state or "created"
        self.label_column = label_column
        self.targets = targets
        self.stats = stats or {}
        self.index_keys = index_keys
        self.preview = preview or []
        self.features: list[Feature] = features or []
        self.run_uri = run_uri
        self.timestamp_key = timestamp_key

    @property
    def targets(self) -> list[DataTarget]:
        """list of material storage targets + their status/path"""
        return self._targets

    @targets.setter
    def targets(self, targets: list[DataTarget]):
        self._targets = ObjectList.from_list(DataTarget, targets)

    def update_target(self, target: DataTarget):
        self._targets.update(target)

    @property
    def features(self) -> list[Feature]:
        """list of features (result of joining features from the source feature sets)"""
        return self._features

    @features.setter
    def features(self, features: list[Feature]):
        self._features = ObjectList.from_list(Feature, features)


class FixedWindowType(Enum):
    CurrentOpenWindow = 1
    LastClosedWindow = 2

    def to_qbk_fixed_window_type(self):
        try:
            from storey import FixedWindowType as QueryByKeyFixedWindowType
        except ImportError as exc:
            raise ImportError("storey not installed, use pip install storey") from exc
        if self == FixedWindowType.LastClosedWindow:
            return QueryByKeyFixedWindowType.LastClosedWindow
        elif self == FixedWindowType.CurrentOpenWindow:
            return QueryByKeyFixedWindowType.CurrentOpenWindow
        else:
            raise NotImplementedError(
                f"Provided fixed window type is not supported. fixed_window_type={self}"
            )


class FeatureVector(ModelObj):
    """Feature vector, specify selected features, their metadata and material views"""

    kind = mlrun.common.schemas.ObjectKind.feature_vector.value
    _dict_fields = ["kind", "metadata", "spec", "status"]

    def __init__(
        self,
        name=None,
        features=None,
        label_feature=None,
        description=None,
        with_indexes=None,
        join_graph: JoinGraph = None,
        relations: typing.Optional[dict[str, dict[str, Union[Entity, str]]]] = None,
    ):
        """Feature vector, specify selected features, their metadata and material views

        example::

            import mlrun.feature_store as fstore

            features = ["quotes.bid", "quotes.asks_sum_5h as asks_5h", "stocks.*"]
            vector = fstore.FeatureVector("my-vec", features)

            # get the vector as a dataframe
            df = vector.get_offline_features().to_dataframe()

            # return an online/real-time feature service
            svc = vector.get_online_feature_service(impute_policy={"*": "$mean"})
            resp = svc.get([{"stock": "GOOG"}])

        :param name:           List of names of targets to delete (default: delete all ingested targets)
        :param features:       list of feature to collect to this vector.
                                Format [<project>/]<feature_set>.<feature_name or `*`> [as <alias>]
        :param label_feature:  feature name to be used as label data
        :param description:    text description of the vector
        :param with_indexes:   whether to keep the entity and timestamp columns in the response
        :param join_graph:     An optional JoinGraph object representing the graph of data joins
                               between feature sets for this feature vector, specified the order and the join types.
        :param relations:      {<feature_set name>: {<column_name>: <other entity object/name>, ...}...}
                               An optional dictionary specifying the relations between feature sets in the
                               feature vector. The keys of the dictionary are feature set names, and the values
                               are dictionaries where the keys represent column names(of the feature set),
                               and the values represent the target entities to join with.
                               The relations provided here will take precedence over the relations that were specified
                               on the feature sets themselves. In case a specific feature set is not mentioned as a key
                               here, the function will fall back to using the default relations defined in the
                               feature set.

        """
        self._spec: FeatureVectorSpec = None
        self._metadata = None
        self._status = None

        self.spec = FeatureVectorSpec(
            description=description,
            features=features,
            label_feature=label_feature,
            with_indexes=with_indexes,
            relations=relations,
            join_graph=join_graph,
        )
        self.metadata = VersionedObjMetadata(name=name)
        self.status = None

        self._entity_df = None
        self._feature_set_fields = {}
        self.feature_set_objects = {}

    @property
    def spec(self) -> FeatureVectorSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", FeatureVectorSpec)

    @property
    def metadata(self) -> VersionedObjMetadata:
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        self._metadata = self._verify_dict(metadata, "metadata", VersionedObjMetadata)

    @property
    def status(self) -> FeatureVectorStatus:
        return self._status

    @status.setter
    def status(self, status):
        self._status = self._verify_dict(status, "status", FeatureVectorStatus)

    @property
    def uri(self):
        """fully qualified feature vector uri"""
        project = self._metadata.project or mlconf.active_project
        name = self._metadata.name

        base = name if not project else f"{project}/{name}"
        uri = get_store_uri(StorePrefix.FeatureVector, base)

        if self._metadata.tag:
            uri += f":{self._metadata.tag}"
        return uri

    def link_analysis(self, name, uri):
        """add a linked file/artifact (chart, data, ..)"""
        self._spec.analysis[name] = uri

    def get_stats_table(self):
        """get feature statistics table (as dataframe)"""
        if self.status.stats:
            feature_aliases = self.get_feature_aliases()
            for old_name, new_name in feature_aliases.items():
                if old_name in self.status.stats:
                    self.status.stats[new_name] = self.status.stats[old_name]
                    del self.status.stats[old_name]
            return pd.DataFrame.from_dict(self.status.stats, orient="index")

    def get_feature_aliases(self):
        feature_aliases = {}
        for feature in self.spec.features:
            column_names = feature.split(" as ")
            # split 'feature_set.old_name as new_name'
            if len(column_names) == 2:
                old_name_with_feature_set, new_name = column_names
                # split 'feature_set.old_name'
                feature_set, old_name = column_names[0].split(".")
                if new_name != old_name:
                    feature_aliases[old_name] = new_name
        return feature_aliases

    def get_target_path(self, name=None):
        target = get_offline_target(self, name=name)
        if target:
            return target.path

    def to_dataframe(self, df_module=None, target_name=None):
        """return feature vector (offline) data as dataframe"""
        driver = get_offline_target(self, name=target_name)
        if not driver:
            raise mlrun.errors.MLRunNotFoundError(
                "there are no offline targets for this feature vector"
            )
        return driver.as_df(df_module=df_module)

    def save(self, tag="", versioned=False):
        """save to mlrun db"""
        db = mlrun.get_run_db()
        self.metadata.project = self.metadata.project or mlconf.active_project
        tag = tag or self.metadata.tag
        as_dict = self.to_dict()
        db.store_feature_vector(as_dict, tag=tag, versioned=versioned)

    def reload(self, update_spec=True):
        """reload/sync the feature set status and spec from the DB"""
        from_db = mlrun.get_run_db().get_feature_vector(
            self.metadata.name, self.metadata.project, self.metadata.tag
        )
        self.status = from_db.status
        if update_spec:
            self.spec = from_db.spec

    def parse_features(self, offline=True, update_stats=False):
        """parse and validate feature list (from vector) and add metadata from feature sets

        :returns
            feature_set_objects: cache of used feature set objects
            feature_set_fields:  list of field (name, alias) per featureset
        """
        processed_features = {}  # dict of name to (featureset, feature object)
        feature_set_objects = self.feature_set_objects or {}
        index_keys = []
        feature_set_fields = collections.defaultdict(list)
        features = copy(self.spec.features)
        label_column_name = None
        label_column_fset = None
        if offline and self.spec.label_feature:
            features.append(self.spec.label_feature)
            feature_set, name, _ = parse_feature_string(self.spec.label_feature)
            self.status.label_column = name
            label_column_name = name
            label_column_fset = feature_set

        def add_feature(name, alias, feature_set_object, feature_set_full_name):
            if alias in processed_features.keys():
                logging.log(
                    logging.WARN,
                    f"feature name/alias {alias} already specified,"
                    " you need to use another alias (feature-set.name [as alias])"
                    f" by default it changed to be {alias}_{feature_set_full_name}",
                )
                alias = f"{alias}_{feature_set_full_name}"

            feature = feature_set_object[name]
            processed_features[alias or name] = (feature_set_object, feature)
            feature_set_fields[feature_set_full_name].append((name, alias))

        for feature in features:
            project_name, feature = parse_project_name_from_feature_string(feature)
            feature_set, feature_name, alias = parse_feature_string(feature)
            if feature_set not in feature_set_objects.keys():
                feature_set_objects[feature_set] = get_feature_set_by_uri(
                    feature_set,
                    project_name if project_name is not None else self.metadata.project,
                )
            feature_set_object = feature_set_objects[feature_set]

            feature_fields = feature_set_object.spec.features.keys()
            if feature_name == "*":
                for field in feature_fields:
                    if field != feature_set_object.spec.timestamp_key and not (
                        feature_set == label_column_fset and field == label_column_name
                    ):
                        if alias:
                            add_feature(
                                field,
                                alias + "_" + field,
                                feature_set_object,
                                feature_set,
                            )
                        else:
                            add_feature(field, field, feature_set_object, feature_set)
            else:
                if feature_name not in feature_fields:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"feature {feature} not found in feature set {feature_set}"
                    )
                add_feature(feature_name, alias, feature_set_object, feature_set)

        for feature_set_name, fields in feature_set_fields.items():
            feature_set = feature_set_objects[feature_set_name]
            for key in feature_set.spec.entities.keys():
                if key not in index_keys:
                    index_keys.append(key)
            for name, alias in fields:
                if name in feature_set.status.stats and update_stats:
                    self.status.stats[name] = feature_set.status.stats[name]
                if name in feature_set.spec.features.keys():
                    feature = feature_set.spec.features[name].copy()
                    feature.origin = f"{feature_set.fullname}.{name}"
                    feature.name = alias or name
                    self.status.features[alias or name] = feature

        self.status.index_keys = index_keys
        return feature_set_objects, feature_set_fields

    def get_feature_set_relations(self, feature_set: Union[str, FeatureSet]):
        if isinstance(feature_set, str):
            feature_set = get_feature_set_by_uri(
                feature_set,
                self.metadata.project,
            )
        name = feature_set.metadata.name
        feature_set_relations = feature_set.spec.relations or {}
        if self.spec.relations and name in self.spec.relations:
            feature_set_relations = self.spec.relations[name]
        return feature_set_relations

    def get_offline_features(
        self,
        entity_rows=None,
        entity_timestamp_column: typing.Optional[str] = None,
        target: DataTargetBase = None,
        run_config: RunConfig = None,
        drop_columns: typing.Optional[list[str]] = None,
        start_time: typing.Optional[Union[str, datetime]] = None,
        end_time: typing.Optional[Union[str, datetime]] = None,
        with_indexes: bool = False,
        update_stats: bool = True,
        engine: typing.Optional[str] = None,
        engine_args: typing.Optional[dict] = None,
        query: typing.Optional[str] = None,
        order_by: typing.Optional[Union[str, list[str]]] = None,
        spark_service: typing.Optional[str] = None,
        timestamp_for_filtering: typing.Optional[Union[str, dict[str, str]]] = None,
        additional_filters: typing.Optional[list] = None,
    ):
        """retrieve offline feature vector results

        specify a feature vector object/uri and retrieve the desired features, their metadata
        and statistics. returns :py:class:`~mlrun.feature_store.OfflineVectorResponse`,
        results can be returned as a dataframe or written to a target

        The start_time and end_time attributes allow filtering the data to a given time range, they accept
        string values or pandas `Timestamp` objects, string values can also be relative, for example:
        "now", "now - 1d2h", "now+5m", where a valid pandas Timedelta string follows the verb "now",
        for time alignment you can use the verb "floor" e.g. "now -1d floor 1H" will align the time to the last hour
        (the floor string is passed to pandas.Timestamp.floor(), can use D, H, T, S for day, hour, min, sec alignment).
        Another option to filter the data is by the `query` argument - can be seen in the example.
        example::

            features = [
                "stock-quotes.bid",
                "stock-quotes.asks_sum_5h",
                "stock-quotes.ask as mycol",
                "stocks.*",
            ]
            vector = FeatureVector(features=features)
            vector.get_offline_features(entity_rows=trades, entity_timestamp_column="time", query="ticker in ['GOOG']
              and bid>100")
            print(resp.to_dataframe())
            print(vector.get_stats_table())
            resp.to_parquet("./out.parquet")

        :param entity_rows:             dataframe with entity rows to join with
        :param target:                  where to write the results to
        :param drop_columns:            list of columns to drop from the final result
        :param entity_timestamp_column: timestamp column name in the entity rows dataframe. can be specified
                                        only if param entity_rows was specified.
        :param run_config:              function and/or run configuration
                                        see :py:class:`~mlrun.feature_store.RunConfig`
        :param start_time:              datetime, low limit of time needed to be filtered. Optional.
        :param end_time:                datetime, high limit of time needed to be filtered. Optional.
        :param with_indexes:            Return vector with/without the entities and the timestamp_key of the feature
                                        sets and with/without entity_timestamp_column and timestamp_for_filtering
                                        columns. This property can be specified also in the feature vector spec
                                        (feature_vector.spec.with_indexes)
                                        (default False)
        :param update_stats:            When set to True (default), updates feature statistics from the requested
                                        feature sets on the vector, which requires 'update' permissions. When set to
                                        False, uses read-only operations that only require 'read' permissions.
        :param engine:                  processing engine kind ("local", "dask", or "spark")
        :param engine_args:             kwargs for the processing engine
        :param query:                   The query string used to filter rows on the output
        :param spark_service:           Name of the spark service to be used (when using a remote-spark runtime)
        :param order_by:                Name or list of names to order by. The name or the names in the list can be the
                                        feature name or the alias of the feature you pass in the feature list.
        :param timestamp_for_filtering: name of the column to filter by, can be str for all the feature sets or a
                                        dictionary ({<feature set name>: <timestamp column name>, ...})
                                        that indicates the timestamp column name for each feature set. Optional.
                                        By default, the filter executes on the timestamp_key of each feature set.
                                        Note: the time filtering is performed on each feature set before the
                                        merge process using start_time and end_time params.
        :param additional_filters: List of additional_filter conditions as tuples.
                            Each tuple should be in the format (column_name, operator, value).
                            Supported operators: "=", ">=", "<=", ">", "<".
                            Example: [("Product", "=", "Computer")]
                            For all supported filters, please see:
                            https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetDataset.html

        """
        if entity_rows is None and entity_timestamp_column is not None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "entity_timestamp_column param "
                "can not be specified without entity_rows param"
            )

        if isinstance(target, BaseStoreTarget) and not target.support_pandas:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"get_offline_features does not support targets that do not support pandas engine."
                f" Target kind: {target.kind}"
            )

        if update_stats:
            feature_vector = _features_to_vector_and_check_permissions(
                self, update_stats
            )
        else:
            feature_vector = self
            verify_feature_vector_permissions(
                feature_vector, mlrun.common.schemas.AuthorizationAction.read
            )

        entity_timestamp_column = (
            entity_timestamp_column or feature_vector.spec.timestamp_field
        )

        merger_engine = get_merger(engine)

        if run_config and not run_config.local:
            return run_merge_job(
                feature_vector,
                target,
                merger_engine,
                engine,
                engine_args,
                spark_service,
                entity_rows,
                entity_timestamp_column=entity_timestamp_column,
                run_config=run_config,
                drop_columns=drop_columns,
                with_indexes=with_indexes,
                query=query,
                order_by=order_by,
                start_time=start_time,
                end_time=end_time,
                timestamp_for_filtering=timestamp_for_filtering,
                additional_filters=additional_filters,
            )

        merger = merger_engine(feature_vector, **(engine_args or {}))
        return merger.start(
            entity_rows,
            entity_timestamp_column,
            target=target,
            drop_columns=drop_columns,
            start_time=start_time,
            end_time=end_time,
            timestamp_for_filtering=timestamp_for_filtering,
            with_indexes=with_indexes,
            update_stats=update_stats,
            query=query,
            order_by=order_by,
            additional_filters=additional_filters,
        )

    def get_online_feature_service(
        self,
        run_config: RunConfig = None,
        fixed_window_type: FixedWindowType = FixedWindowType.LastClosedWindow,
        impute_policy: typing.Optional[dict] = None,
        update_stats: bool = False,
        entity_keys: typing.Optional[list[str]] = None,
    ) -> OnlineVectorService:
        """initialize and return online feature vector service api,
        returns :py:class:`~mlrun.feature_store.OnlineVectorService`

        :**usage**:
            There are two ways to use the function:

            1. As context manager

                Example::

                    with vector_uri.get_online_feature_service() as svc:
                        resp = svc.get([{"ticker": "GOOG"}, {"ticker": "MSFT"}])
                        print(resp)
                        resp = svc.get([{"ticker": "AAPL"}], as_list=True)
                        print(resp)

                Example with imputing::

                    with vector_uri.get_online_feature_service(entity_keys=['id'],
                                                    impute_policy={"*": "$mean", "amount": 0)) as svc:
                        resp = svc.get([{"id": "C123487"}])

            2. as simple function, note that in that option you need to close the session.

                Example::

                    svc = vector_uri.get_online_feature_service(entity_keys=["ticker"])
                    try:
                        resp = svc.get([{"ticker": "GOOG"}, {"ticker": "MSFT"}])
                        print(resp)
                        resp = svc.get([{"ticker": "AAPL"}], as_list=True)
                        print(resp)

                    finally:
                        svc.close()

                Example with imputing::

                    svc = vector_uri.get_online_feature_service(entity_keys=['id'],
                                                    impute_policy={"*": "$mean", "amount": 0))
                    try:
                        resp = svc.get([{"id": "C123487"}])
                    except Exception as e:
                        handling exception...
                    finally:
                        svc.close()

        :param run_config:          function and/or run configuration for remote jobs/services
        :param impute_policy:       a dict with `impute_policy` per feature, the dict key is the feature name and the
                                    dict value indicate which value will be used in case the feature is NaN/empty, the
                                    replaced value can be fixed number for constants or $mean, $max, $min, $std, $count
                                    for statistical values.
                                    "*" is used to specify the default for all features, example: `{"*": "$mean"}`
        :param fixed_window_type:   determines how to query the fixed window values which were previously inserted by
                                    ingest
        :param update_stats:        update features statistics from the requested feature sets on the vector.
                                    Default: False.
        :param entity_keys:         Entity list of the first feature_set in the vector.
                                    The indexes that are used to query the online service.
        :return:                    Initialize the `OnlineVectorService`.
                                    Will be used in subclasses where `support_online=True`.
        """
        feature_vector = _features_to_vector_and_check_permissions(self, True)

        engine_args = {"impute_policy": impute_policy}
        merger_engine = get_merger("storey")
        # todo: support remote service (using remote nuclio/mlrun function if run_config)

        merger = merger_engine(feature_vector, **engine_args)

        return merger.init_online_vector_service(
            entity_keys, fixed_window_type, update_stats=True
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Pyspark example called by mlrun_spark_k8s.ipynb


from pyspark.sql import SparkSession

from mlrun import get_or_create_ctx

# Acquire MLRun context
mlctx = get_or_create_ctx("spark-function")

# Get MLRun parameters
mlctx.logger.info("!@!@!@!@!@ Getting env variables")
READ_OPTIONS = mlctx.get_param("data_sources")
QUERY = mlctx.get_param("query")
WRITE_OPTIONS = mlctx.get_param("write_options")

# Create spark session
spark = SparkSession.builder.appName("Spark function").getOrCreate()


# Loading data from a JDBC source
for data_source in READ_OPTIONS:
    spark.read.load(**READ_OPTIONS[data_source]).createOrReplaceTempView(data_source)

# Transform the data using SQL query
spark.sql(QUERY).write.save(**WRITE_OPTIONS)

# write the result datadrame to destination
mlctx.logger.info("!@!@!@!@!@ Saved")
spark.stop()



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from datetime import datetime
from typing import Optional

from pydantic.v1 import BaseModel


class FunctionsType(enum.Enum):
    APPLICATION = "application"
    INFRA = "infra"


class FunctionSummary(BaseModel):
    """
    Function summary model. Includes metadata about the function, such as its name, as well as statistical
    metrics such as the number of detections and possible detections. A function summary can be from either a
    model monitoring application (type "application") or an infrastructure function (type "infra").
    """

    type: FunctionsType
    name: str
    application_class: str
    project_name: str
    updated_time: datetime
    status: Optional[str] = None
    base_period: Optional[int] = None
    stats: Optional[dict] = None

    @classmethod
    def from_function_dict(
        cls,
        func_dict: dict,
        func_type=FunctionsType.APPLICATION,
        base_period: Optional[int] = None,
        stats: Optional[dict] = None,
    ):
        """
        Create a FunctionSummary instance from a dictionary.
        """

        return cls(
            type=func_type,
            name=func_dict["metadata"]["name"]
            if func_type != FunctionsType.APPLICATION
            else func_dict["spec"]
            .get("graph", {})
            .get("steps", {})
            .get("PrepareMonitoringEvent", {})
            .get("class_args", {})
            .get("application_name"),
            application_class=""
            if func_type != FunctionsType.APPLICATION
            else func_dict["spec"]
            .get("graph", {})
            .get("steps", {})
            .get("PushToMonitoringWriter", {})
            .get("after", [None])[0],
            project_name=func_dict["metadata"]["project"],
            updated_time=func_dict["metadata"].get("updated"),
            status=func_dict["status"].get("state"),
            base_period=base_period,
            stats=stats or {},
        )
